/******************************************************************************
 * program:     wp2latex                                                      *
 * function:    Convert WordPerfect files into LaTeX.			      *
 * modul:       pass1wmf.cc                                                   *
 * description: Opens Windows MetaFile.					      *
 * licency:     GPL		                                              *
 ******************************************************************************/
// See https://en.wikipedia.org/wiki/Windows_Metafile
//     http://www.fileformat.info/format/wmf/egff.htm
// for WMF description.
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

//Atoms library
#include <strings.h>

#include "wp2latex.h"
#include "images/raster.h"
#include "images.h"


void WMF_STRETCHDIB(Image &Img, FILE *f, DWORD ParamFilePos);
int SavePictureEPS(const char *Name,const Image &Img);
string CutFileName(const char *FullFilename); //from wp2lfuti.cc



typedef  enum  
{ 
  MM_TEXT = 0x0001, 
  MM_LOMETRIC = 0x0002, 
  MM_HIMETRIC = 0x0003, 
  MM_LOENGLISH = 0x0004, 
  MM_HIENGLISH = 0x0005, 
  MM_TWIPS = 0x0006, 
  MM_ISOTROPIC = 0x0007, 
  MM_ANISOTROPIC = 0x0008 
} TMapMode;

/// Get scalling factor for a given map mode.
float GetScale2PSU(TMapMode MapM)
{
  switch(MapM)
  {
    case MM_TEXT:	return 1; 
    case MM_LOMETRIC:   return mm2PSu(0.1f);
    case MM_HIMETRIC:   return mm2PSu(0.01f);
    case MM_LOENGLISH:  return mm2PSu(0.254f);
    case MM_HIENGLISH:  return mm2PSu(0.0254f);
    case MM_TWIPS:	return mm2PSu(0.017638889f);
    case MM_ISOTROPIC:
    case MM_ANISOTROPIC: break;
  }
  return 1;
}
 

typedef struct _StandardMetaRecord
{
    DWORD Size;          /**< Total size of the record in WORDs */
    WORD  Function;      /**< Function number (defined in WINDOWS.H) */
    DWORD ParamFilePos;
    //WORD  Parameters[];  /**< Parameter values passed to function */
} WMFRECORD;



class TconvertedPass1_WMF: public TconvertedPass1
     {
protected:
     int NoConvertImage;

public:
     TconvertedPass1_WMF(void);

     virtual int Convert_first_pass(void);
     virtual int Dispatch(int FuncNo, ...);

     SWORD YOffset, XOffset;
     SWORD YExtent, XExtent;

     SWORD PositionX, PositionY;
     Image Img;     
     WORD MapMode;
     float FontSize;
     FloatBBox bbx;
     PS_State PSS;
     string PSData;
     _StandardMetaRecord WmfRec;     

protected:
     bool CheckWmfRecSz(WORD WSizeRequest);

     void parse_CreateBrushIndirect(void);
     void parse_Ellipse(void);
     void parse_FontIndirect(void);
     void parse_Chord(void);
     void parse_LineTo(void);
     void parse_MoveTo(void);
     void parse_Polygon(void);
     void parse_PolyPolygon(void);
     void parse_PolyLine(void);     
     void wmf_ESCAPE(void);
     void parse_ExtTextOut(void);
     void parse_SetBkColor(void);
     void parse_SetMapMode(void);
     void parse_SetTextColor(void);
     void parse_SetWindowExt(void);
     void parse_SetWindowOrg(void);
     void parse_TextOut(void);
     void StretchDIBits(void);
     };


TconvertedPass1_WMF::TconvertedPass1_WMF(void): NoConvertImage(0)
{
  PositionX = PositionY = 0;
  YOffset = XOffset = 0;
  YExtent = XExtent = 0;
  bbx.MinX=65537; bbx.MaxX=-32767; bbx.MinY=65537; bbx.MaxY=-32767;
  memset(&PSS, 0, sizeof(PSS));
  PSS.FillPattern = 1;
  FontSize = 1100;		// Default font size
  MapMode = MM_ANISOTROPIC;
}


/*Register translators here*/
TconvertedPass1 *Factory_WMF(void) {return new TconvertedPass1_WMF;}
FFormatTranslator FormatWMFWrapper("WMF",Factory_WMF);
FFormatTranslator FormatEMFWrapper("EMF",Factory_WMF);

typedef struct _WindowsMetaHeader
{
  WORD  FileType;       ///< Type of metafile (0=memory, 1=disk).
  WORD  HeaderSize;     ///< Size of header in WORDS (always 9).
  WORD  Version;        ///< Version of Microsoft Windows used.
  DWORD FileSize;       ///< Total size of the metafile in WORDs.
  WORD  NumOfObjects;   ///< Number of objects in the file.
  DWORD MaxRecordSize;  ///< The size of largest record in WORDs.
  WORD  NumOfParams;    ///< Not Used (always 0).
} WMFHEAD;


inline long LoadWMFHeader(FILE *f, _WindowsMetaHeader & SU)
{
#if defined(__PackedStructures__)
return(fread(&SU,1,sizeof(_WindowsMetaHeader),f));
#else
return(loadstruct(f,"wwwdwdw",&SU.FileType,&SU.HeaderSize,
	&SU.Version,&SU.FileSize,&SU.NumOfObjects,&SU.MaxRecordSize,&SU.NumOfParams));
#endif
}


/** Load header for one WMF record. */
bool LoadWmfRecord(FILE *f, _StandardMetaRecord &WmfRec)
{
  Rd_dword(f,&WmfRec.Size);
  if(feof(f) || WmfRec.Size<3) return false;
  Rd_word(f,&WmfRec.Function);
  WmfRec.ParamFilePos = ftell(f);
  return true;  
}


void TconvertedPass1_WMF::wmf_ESCAPE(void)
{
static WORD MathType[] = {15,12,'M'+256*'a','t'+256*'h','T'+256*'y','p'+256*'e',0};
int Size;
WORD w;

  if(WmfRec.Size-3<sizeof(MathType)/sizeof(WORD)) goto NoMTEF;

  fseek(wpd,WmfRec.ParamFilePos,SEEK_SET);
  for(Size=0; Size<sizeof(MathType)/sizeof(WORD); Size++)
      {
      Rd_word(wpd,&w);
      if(w!=MathType[Size]) goto NoMTEF;
      }

  strcpy(ObjType+6," MTEF");
  if(err != NULL)
    {
    perc.Hide();
    fprintf(err, _("\nNote: MTEF formulas inside WMF are not converted yet."));
    }

  strcpy(ObjType,"!Escape - MTEF");
  return;


NoMTEF:
  strcpy(ObjType,"!Escape");
}


/** This callback from parent allows to provide some generic functionality. */
int TconvertedPass1_WMF::Dispatch(int FuncNo, ...)
{
#ifdef DEBUG
  fprintf(log,"\n#TconvertedPass1_WMF::Dispatch(%d) ",FuncNo);fflush(log);
#endif
va_list ap;

 va_start(ap, FuncNo); 
 switch(FuncNo)
   {
   case DISP_EXTRACTIMAGE:
         {
	 Image **FillME = va_arg(ap,Image **);
         if(FillME) *FillME = new Image(Img);
         va_end(ap);
	 return 0;
	 }
   case DISP_NOCONVERTIMAGE:
	 NoConvertImage = va_arg(ap,int);
         va_end(ap);
         return 0;
   } 
  va_end(ap);

return(-1);
}


/** Check whether WMF record has sufficient capacity.
 * @param[in]	WSizeRequest	Capacity of data payload in WORDs, without header record. */
bool TconvertedPass1_WMF::CheckWmfRecSz(WORD WSizeRequest)
{
  if(WSizeRequest > WmfRec.Size-3)	// {RecordSize}[RecordFunction]
  {
    if(err != NULL)
        fprintf(err,_("\nError: WMF record %X size %d is too small, expected %d!"), 
                WmfRec.ParamFilePos ,WmfRec.Size ,WSizeRequest+2);
    return false;
  }
  return true;
}


void TconvertedPass1_WMF::parse_Ellipse(void)
{
SWORD	BottomRect, TopRect, RightRect, LeftRect;

  if(!CheckWmfRecSz(4))
  {
    strcpy(ObjType,"!Ellipse");
    return;
  }

  Rd_word(wpd, (WORD*)&BottomRect);
  Rd_word(wpd, (WORD*)&RightRect);
  Rd_word(wpd, (WORD*)&TopRect);
  Rd_word(wpd, (WORD*)&LeftRect);

  UpdateBBox(bbx, 0, LeftRect, BottomRect, RightRect-LeftRect, TopRect-BottomRect);

  const float Scale = GetScale2PSU((TMapMode)MapMode);
  PSData += "\nnewpath";
  PSData.cat_printf("\n%2.2f %2.2f %2.2f %2.2f %d %d DrawEllipse",
			 Scale*(RightRect+LeftRect)/2, Scale*(TopRect+BottomRect)/2,
			 Scale*labs(RightRect-LeftRect)/2, Scale*labs(TopRect-BottomRect)/2,
			 0, 360);
  PSData+="\nstroke";

strcpy(ObjType,"Ellipse");
}


void TconvertedPass1_WMF::parse_Chord(void)
{
SWORD	XRadial1, YRadial1, XRadial2, YRadial2;
SWORD	BottomRect, RightRect, TopRect, LeftRect;

  if(!CheckWmfRecSz(4))
  {
    strcpy(ObjType,"!Chord");
    return;
  }

  Rd_word(wpd, (WORD*)&XRadial2);
  Rd_word(wpd, (WORD*)&YRadial2);
  Rd_word(wpd, (WORD*)&XRadial1);
  Rd_word(wpd, (WORD*)&YRadial1);
  Rd_word(wpd, (WORD*)&BottomRect);
  Rd_word(wpd, (WORD*)&RightRect);
  Rd_word(wpd, (WORD*)&TopRect);
  Rd_word(wpd, (WORD*)&LeftRect);

strcpy(ObjType,"!Chord");	// @TODO: dopsat!
}


void TconvertedPass1_WMF::parse_FontIndirect(void)
{
SWORD Height;
WORD Width, Escapement, Orientation, Weight;
BYTE Italic, Underline, Strikeout, Charset, OutPrecision, ClipPrecision, Quality, PitchAndFamily;

  Rd_word(wpd, (WORD*)&Height);
  Rd_word(wpd, &Width);
  Rd_word(wpd, &Escapement);
  Rd_word(wpd, &Orientation);
  Rd_word(wpd, &Weight);

  if(Height!=0) FontSize = labs(Height) * 50.0;

strcpy(ObjType,"CreateFontIndirect");
}


void TconvertedPass1_WMF::parse_MoveTo(void)
{
  if(CheckWmfRecSz(2))
    {
    Rd_word(wpd, (WORD*)&PositionX);
    Rd_word(wpd, (WORD*)&PositionY);
    }

  strcpy(ObjType,"MoveTo");  
}


void TconvertedPass1_WMF::parse_CreateBrushIndirect(void)
{
WORD dummy;
  if(CheckWmfRecSz(3))
    {
    Rd_word(wpd, &dummy);

    PSS.FillColor.Red   = fgetc(wpd);
    PSS.FillColor.Green = fgetc(wpd);
    PSS.FillColor.Blue  = fgetc(wpd);
    }

  strcpy(ObjType,"CreateBrushIndirect");
}


void TconvertedPass1_WMF::parse_SetBkColor(void)
{
  if(CheckWmfRecSz(2))
    {
    PSS.FillBackground.Red   = fgetc(wpd);
    PSS.FillBackground.Green = fgetc(wpd);
    PSS.FillBackground.Blue  = fgetc(wpd);
    }

  strcpy(ObjType,"SetBkColor");
}


void TconvertedPass1_WMF::parse_SetTextColor(void)
{
  if(CheckWmfRecSz(2))
    {
    PSS.TextColor.Red   = fgetc(wpd);
    PSS.TextColor.Green = fgetc(wpd);
    PSS.TextColor.Blue  = fgetc(wpd);
    }

  strcpy(ObjType,"SetTextColor");
}


void TconvertedPass1_WMF::parse_SetWindowExt(void)
{
  Rd_word(wpd, (WORD*)&YExtent);
  Rd_word(wpd, (WORD*)&XExtent);
  strcpy(ObjType,"SetWindowExt");
}


void TconvertedPass1_WMF::parse_SetWindowOrg(void)
{
  Rd_word(wpd, (WORD*)&YOffset);
  Rd_word(wpd, (WORD*)&XOffset);
  strcpy(ObjType,"SetWindowOrg");
}


void TconvertedPass1_WMF::parse_LineTo(void)
{
SWORD LineX, LineY;
  if(CheckWmfRecSz(2))
    {
    Rd_word(wpd, (WORD*)&LineX);
    Rd_word(wpd, (WORD*)&LineY);

    UpdateBBox(bbx, 0, PositionX, PositionY, LineX-PositionX, LineY-PositionY);

      {
      const float Points[4] = {
                   PositionX,
                   PositionY,
                   LineX,
                   LineY};
      Line2PS(Points, 2, PSData, &PSS, 0, GetScale2PSU((TMapMode)MapMode));
      }
    }
  strcpy(ObjType,"LineTo");  
}


void TconvertedPass1_WMF::parse_Polygon(void)
{
  WORD PolySize;  

  if(!CheckWmfRecSz(1))
  {
POLY_FAIL:
    strcpy(ObjType,"!Polygon");
    return;
  }

  Rd_word(wpd, &PolySize);
  if(PolySize > (WmfRec.Size - 3 - 1)/2)
    PolySize = (WmfRec.Size - 3 - 1)/2;
  float *Points = LoadPoints(this, PolySize, bbx, NULL);
  if(Points==NULL)  goto POLY_FAIL;  

  Polygon2PS(Points, PolySize, PSData, &PSS, true, GetScale2PSU((TMapMode)MapMode));

  free(Points);
  strcpy(ObjType,"Polygon");
  return;
}

void TconvertedPass1_WMF::parse_PolyPolygon(void)
{
WORD PolyCount;
WORD *PolySizes;
WORD i;

  if(!CheckWmfRecSz(1+1))
  {
POLY_FAIL:
    strcpy(ObjType,"!PolyPolygon");
    return;
  }

  Rd_word(wpd, &PolyCount);
  if(PolyCount<=0) goto POLY_FAIL;

  PolySizes = (WORD*)malloc(PolyCount*sizeof(WORD));
  if(PolySizes==NULL) goto POLY_FAIL;
  for(i=0; i<PolyCount;i++)
  {
    Rd_word(wpd, &PolySizes[i]);
  }

  for(i=0; i<PolyCount;i++)
  {
    if(PolySizes[i]<=0) continue;
    //if(PolySize > (WmfRec.Size - 3 - 1)/2)
    //  PolySize = (WmfRec.Size - 3 - 1)/2;
    float *Points = LoadPoints(this, PolySizes[i], bbx, NULL);
    if(Points==NULL)
        {free(PolySizes);goto POLY_FAIL;}

    Polygon2PS(Points, PolySizes[i], PSData, &PSS, true, GetScale2PSU((TMapMode)MapMode));
    free(Points);
  }
  free(PolySizes);
  strcpy(ObjType,"PolyPolygon");
  return;
}


void TconvertedPass1_WMF::parse_PolyLine(void)
{
  WORD PolySize;  

  if(!CheckWmfRecSz(1))
  {
POLY_FAIL:
    strcpy(ObjType,"!Polyline");
    return;
  }

  Rd_word(wpd, &PolySize);
  if(PolySize > (WmfRec.Size - 3 - 1)/2)
    PolySize = (WmfRec.Size - 3 - 1)/2;  
  float *Points = LoadPoints(this, PolySize, bbx, NULL);
  if(Points==NULL)  goto POLY_FAIL;

  Line2PS(Points, PolySize, PSData, &PSS, false, GetScale2PSU((TMapMode)MapMode));

  free(Points);
  strcpy(ObjType,"Polyline");  
}


void TconvertedPass1_WMF::parse_SetMapMode(void)
{
  Rd_word(wpd, &MapMode);
  strcpy(ObjType,"SetMapMode");
}


void TconvertedPass1_WMF::parse_TextOut(void)
{
WORD StringLen;
SWORD XStart, YStart;

  if(!CheckWmfRecSz(3))
  {
TEXT_OUT_FAIL:
    strcpy(ObjType,"!TextOut");    
    return;
  }

  Rd_word(wpd, &StringLen);
  if(!CheckWmfRecSz((StringLen+1)/2)) goto TEXT_OUT_FAIL;

  string Text;
  XStart = StringLen & 1;
  while(StringLen-->0)
  {
    char ch = fgetc(wpd);
    if(ch==0)
    {
      fseek(wpd,StringLen,SEEK_CUR);
      break;
    }
    if(ch==')' || ch=='(') Text+='\\';
    Text += ch;
  }
  if(XStart>0)			// Word allignment.
    fseek(wpd,1,SEEK_CUR);

  Rd_word(wpd, (WORD*)&YStart);
  Rd_word(wpd, (WORD*)&XStart);
  UpdateBBox(bbx, 0, XStart, YStart, 0, 0);

  const float Scale = GetScale2PSU((TMapMode)MapMode);
  StoreText2PS(PSData, Text, &PSS, 0,
	       Scale*XStart, Scale*YStart, Scale*XStart+12*Text.length(), Scale*YStart+mm2PSu(FontSize/50.0/2.66),
	       FontSize/50.0/2.66);

  strcpy(ObjType,"TextOut");
}


void TconvertedPass1_WMF::parse_ExtTextOut(void)
{
SWORD x_pos, y_pos;
WORD fuOptions, hdc;
char c;

  if(!CheckWmfRecSz(4))
  {
TEXT_OUT_FAIL:
    strcpy(ObjType,"!ExtTextOut");
    return;
  }
  
  Rd_word(wpd, (WORD*)&y_pos);
  Rd_word(wpd, (WORD*)&x_pos);
  Rd_word(wpd, &hdc);  
  Rd_word(wpd, &fuOptions);

  string Text;
  int TextLen = 2*(WmfRec.Size - 3 - 4) / 3;
  while(!feof(wpd))
    {
    if(TextLen-- <= 0) break;
    c = fgetc(wpd);
    if(c==0) break;
    if(c==')' || c=='(') Text+='\\';
    Text += c;
    }
  if(Text.length()==0) goto TEXT_OUT_FAIL;

  UpdateBBox(bbx, 0, x_pos, y_pos, 0, 0);

  const float Scale = GetScale2PSU((TMapMode)MapMode);
  StoreText2PS(PSData, Text, &PSS, 0,
		Scale*x_pos, Scale*y_pos, Scale*x_pos+12*Text.length(), Scale*y_pos+12,
		FontSize/50.0/2.66);

  strcpy(ObjType,"ExtTextOut");
}


void TconvertedPass1_WMF::StretchDIBits(void)
{
Image *pImg = &Img;

  while(pImg->Next!=NULL)
    pImg = pImg->Next;

  if(pImg->Raster==NULL || pImg->Raster->Data2D==NULL)
    WMF_STRETCHDIB(*pImg, wpd, WmfRec.ParamFilePos);
  else
  {
    Image Img2;
    WMF_STRETCHDIB(Img2, wpd, WmfRec.ParamFilePos);
    if(Img2.Raster!=NULL && Img2.Raster->Size1D>0)
    {
      if(Img2.Raster->Size1D==pImg->Raster->Size1D && Img2.Raster->Size2D==1 && 
         Img2.Raster->GetPlanes()==pImg->Raster->GetPlanes())
      {
        void **NewBlock = (void**)realloc(pImg->Raster->Data2D, (pImg->Raster->Size2D+1)*sizeof(void**));
        if(NewBlock!=NULL)
        {
          NewBlock[pImg->Raster->Size2D] = Img2.Raster->Data2D[0];
          Img2.Raster->Data2D[0] = NULL;
	  // Img2.Raster->Size2D must not be schrinked here, schrink will cause leak. Dtor will free empty ptr.
          pImg->Raster->Data2D = NewBlock;
          pImg->Raster->Size2D++;
        }
      }
      else
      {
        pImg->Next = new(Image);
        pImg = pImg->Next;
        pImg->AttachPalette(Img2.Palette);
        pImg->AttachRaster(Img2.Raster);
      }
    }
  }
  strcpy(ObjType,"StretchDIBits");
}


/*******************************************************************/
/* This procedure provides all needed processing for the first pass*/
int TconvertedPass1_WMF::Convert_first_pass(void)
{
#ifdef DEBUG
  fprintf(log,"\n#Convert_pass1_WMF() ");fflush(log);
#endif
_WindowsMetaHeader WmfHead;
int Rec;
int RetVal=0;
DWORD NewPos;
//DWORD fsize=filesize(wpd);

  if(Verbosing >= 1) printf(_("Opening Windows MetaFile:"));

  if(LoadWMFHeader(wpd,WmfHead)!=18) return(-1);
  
  if(WmfHead.HeaderSize!=9)
    {
    if(err != NULL)
        fprintf(err, _("\nError: Invalid WMF header size %d, expected size=9!"), WmfHead.HeaderSize);
    return(-1);
    }

  //perc.Init(ftell(wpd), fsize,_("First pass WMF:") );

  ActualPos = ftell(wpd);
  Rec = 0;  
  while(!feof(wpd))
    {    
    if(LoadWmfRecord(wpd,WmfRec))
      {
      *ObjType = 0;
      NewPos = 0;

      switch(WmfRec.Function)
	{
	case 0x0000: strcpy(ObjType,"WMF End"); break;

	case 0x001E: strcpy(ObjType,"!SaveDC"); break;

	case 0x0035: strcpy(ObjType,"!RealizePalette"); break;
	case 0x0037: strcpy(ObjType,"!SetPalEntries"); break;

	case 0x004F: strcpy(ObjType,"!StartPage"); break;
        case 0x0050: strcpy(ObjType,"!EndPage"); break;

	case 0x0052: strcpy(ObjType,"!AbortDoc"); break;
        case 0x005E: strcpy(ObjType,"!EndDoc"); break;

	case 0x00F7: strcpy(ObjType,"!CreatePalette"); break;
	case 0x00F8: strcpy(ObjType,"!CreateBrush"); break;

	case 0x0102: strcpy(ObjType,"!SetBkMode"); break;
	case 0x0103: parse_SetMapMode(); break;
	case 0x0104: strcpy(ObjType,"!SetROP2"); break;
	case 0x0105: strcpy(ObjType,"!SetRelabs"); break;
	case 0x0106: strcpy(ObjType,"!SetPolyFillMode"); break;
	case 0x0107: strcpy(ObjType,"!SetStretchBltMode"); break;
	case 0x0108: strcpy(ObjType,"!SetTextCharExtra"); break;

	case 0x0127: strcpy(ObjType,"!RestoreDC"); break;
	case 0x012A: strcpy(ObjType,"!InvertRegion"); break;
	case 0x012B: strcpy(ObjType,"!PaintRegion"); break;
	case 0x012C: strcpy(ObjType,"!SelectClipRegion"); break;
	case 0x012D: strcpy(ObjType,"!SelectObject"); break;
	case 0x012E: strcpy(ObjType,"!SetTextAlign"); break;

	case 0x0139: strcpy(ObjType,"!ResizePalette"); break;

	case 0x0142: strcpy(ObjType,"!DibCreatePatternBrush"); break;

	case 0x014C: strcpy(ObjType,"!ResetDc"); break;
	case 0x014D: strcpy(ObjType,"!StartDoc"); break;

	case 0x01F0: strcpy(ObjType,"!DeleteObject"); break;

	case 0x01F9: strcpy(ObjType,"!CreatePatternBrush"); break;

	case 0x0201: parse_SetBkColor(); break;

	case 0x0209: parse_SetTextColor(); break;
	case 0x020A: strcpy(ObjType,"!SetTextJustification"); break;
	case 0x020B: parse_SetWindowOrg(); break;
	case 0x020C: parse_SetWindowExt(); break;
	case 0x020D: strcpy(ObjType,"!SetViewportOrg"); break;
	case 0x020E: strcpy(ObjType,"!SetViewportExt"); break;
	case 0x020F: strcpy(ObjType,"!OffsetWindowOrg"); break;

	case 0x0211: strcpy(ObjType,"!OffsetViewportOrg"); break;
	case 0x0213: parse_LineTo(); break;
	case 0x0214: parse_MoveTo(); break;

	case 0x0220: strcpy(ObjType,"!OffsetClipRgn"); break;

	case 0x0228: strcpy(ObjType,"!FillRegion"); break;

	case 0x0231: strcpy(ObjType,"!SetMapperFlags"); break;

	case 0x0234: strcpy(ObjType,"!SelectPalette"); break;

	case 0x02FA: strcpy(ObjType,"!CreatePenIndirect"); break;
	case 0x02FB: parse_FontIndirect(); break;
	case 0x02FC: parse_CreateBrushIndirect(); break;
	case 0x02FD: strcpy(ObjType,"!CreateBitmapIndirect"); break;

	case 0x0324: parse_Polygon(); break;
	case 0x0325: parse_PolyLine(); break;

	case 0x0410: strcpy(ObjType,"!ScaleWindowExt"); break;

	case 0x0412: strcpy(ObjType,"!ScaleViewportExt"); break;

	case 0x0415: strcpy(ObjType,"!ExcludeClipRect"); break;
	case 0x0416: strcpy(ObjType,"!IntersectClipRect"); break;

	case 0x0418: parse_Ellipse(); break;
	case 0x0419: strcpy(ObjType,"!FloodFill"); break;

	case 0x041B: strcpy(ObjType,"!Rectangle"); break;

	case 0x041F: strcpy(ObjType,"!SetPixel"); break;

	case 0x0429: strcpy(ObjType,"!FrameRegion"); break;

	case 0x0436: strcpy(ObjType,"!AnimatePalette"); break;

	case 0x0521: parse_TextOut(); break;
	case 0x0538: parse_PolyPolygon(); break;

	case 0x0548: strcpy(ObjType,"!ExtFloodFill"); break;

	case 0x626: wmf_ESCAPE(); break;

	case 0x061C: strcpy(ObjType,"!RoundRect"); break;
	case 0x061D: strcpy(ObjType,"!PatBlt"); break;

	case 0x062F: strcpy(ObjType,"!DrawText"); break;

	case 0x06FE: strcpy(ObjType,"!CreateBitmap"); break;
	case 0x06FF: strcpy(ObjType,"!CreateRegion"); break;

        case 0x0817: strcpy(ObjType,"!Arc"); break;

	case 0x081A: strcpy(ObjType,"!Pie"); break;

 	case 0x0830: parse_Chord(); break;

	case 0x0922: strcpy(ObjType,"!BitBlt"); break;

	case 0x0940: strcpy(ObjType,"!DibBitblt"); break;

	case 0x0A32: parse_ExtTextOut(); break;

	case 0x0B23: strcpy(ObjType,"!StretchBlt"); break;

	case 0x0B41: strcpy(ObjType,"!DibStretchBlt"); break;

	case 0x0d33: strcpy(ObjType,"!SetDibToDev"); break;

	case 0x0F43: StretchDIBits(); break;
		     break;
	}

      if(log!=NULL)
	{   /**/
	fprintf(log,_("\n%*s{WMF function:%3Xh; size:%4lu; pos:%lXh}"),
		  recursion*2,"",WmfRec.Function,2*WmfRec.Size,ActualPos);
	if(*ObjType!=0) fprintf(log," %s",ObjType);

	fprintf(log,"{");
	fseek(wpd,WmfRec.ParamFilePos,SEEK_SET);
	for(int i=0; i<WmfRec.Size-3; i++)
	  //fprintf(log,"%X ",WmfRec->Parameters[i]);
	  {
	  WORD w;
          Rd_word(wpd,&w);
	  if(feof(wpd)) break;
	  int j=w%256;
	  fprintf(log,"%X",j);
	  if(j>=' ') fprintf(log,"(%c)",j);
	  j=w/256;
	  fprintf(log,",%X",j);
	  if(j>=' ') fprintf(log,"(%c)",j);
	  fprintf(log,",");

	  if(i>100) {fprintf(log," ... ");break;}
	  }
	fprintf(log,"}");
	}
      if(*ObjType==0) UnknownObjects++;

      if(NewPos!=0) ActualPos = NewPos;
	       else ActualPos+= 2*WmfRec.Size;      

      fseek(wpd,ActualPos,SEEK_SET);
      }
    else
      ActualPos=ftell(wpd);

    Rec++;
    }

   if(Img.dx!=0 && Img.dy!=0 && Img.Raster!=NULL)
   {
     if(Img.dy<0)
     {        
       Img.y += Img.dy;
       Img.dy = -Img.dy;
     }
     else
       Flip2D(Img.Raster);

     if(Img.dx<0)
     {        
       Img.x += Img.dx;
       Img.dx = -Img.dx;
       Flip1D(Img.Raster);
     }     
   }

  if(PSData.length()>0 && Img.Comment==NULL)
  {
    const float Scale = GetScale2PSU((TMapMode)MapMode) * 25.4f / 71.0f;	// convert PSu to WPGu (quite bad).
    Img.Comment = PSData.ExtractString();
    if(Img.dx!=0 && Img.dy!=0 && Img.Raster!=NULL)
    {
      if(Img.Comment!=NULL)	// Move raster data to different image frame.
      {
        Image *Img2 = &Img;
        while(Img2->Next!=NULL)
          Img2 = Img2->Next;
        Img2->Next = new Image();
        Img2 = Img2->Next;

        Img2->x =  bbx.MinX * Scale;
        Img2->y =  bbx.MinY * Scale;
        Img2->dx = (bbx.MaxX - bbx.MinX) * Scale;
        Img2->dy = (bbx.MaxY - bbx.MinY) * Scale;
        Img2->Comment = Img.Comment; Img.Comment=NULL;
      }
    }
    else	// Use whole frame as bounding box.
    {      
      Img.x =  bbx.MinX;
      Img.y =  bbx.MinY;
      Img.dx = bbx.MaxX - bbx.MinX;
      Img.dy = bbx.MaxY - bbx.MinY;  
    }
    Img.x *= Scale;
    Img.y *= Scale;
    Img.dx *= Scale;
    Img.dy *= Scale;
  }

  if(!NoConvertImage && (Img.Raster!=NULL || Img.Comment!=NULL))
    {
    for(Image *pImg=&Img; pImg!=NULL; pImg=pImg->Next)
      if(pImg->Raster!=NULL)
        ReducePalette(pImg,256);

    string NewFilename = MergePaths(OutputDir,RelativeFigDir) + GetFullFileName(GetSomeImgName(".eps"));
    if(SavePictureEPS(NewFilename(),Img)<0)
	{
        if(err != NULL)
	  {
	  perc.Hide();
	  fprintf(err, _("\nError: Cannot save file: %s !"),NewFilename());
	  }
	return 0;
        }

    NewFilename = CutFileName(NewFilename); 	//New Filename only

    if(InputPS & IMG_graphicx)	//graphicx.sty
      {
      fprintf(strip," \\begin{center}");
      fprintf(strip,"\\includegraphics");
      fprintf(strip,"[width=\\textwidth]");
      fprintf(strip,"{\\FigDir/%s.eps}",NewFilename() );
      fprintf(strip," \\end{center}");
      putc('\n',strip);  rownum++;
    }
    else if(InputPS & IMG_epsfig)	//epsfig.sty
      {
      fprintf(strip," \\begin{center}");
      fprintf(strip,"\\epsfig{file=\\FigDir/%s.eps,",NewFilename() );
      fprintf(strip,"width=\\textwidth");
      putc('}',strip);
      fprintf(strip," \\end{center}");
      putc('\n',strip);  rownum++;
      }
    else if(InputPS & IMG_graphics)	//graphics.sty
      {
      fprintf(strip," \\begin{center}");
      fprintf(strip,"\\includegraphics");
      fprintf(strip,"{\\FigDir/%s.eps}",NewFilename() );
      fprintf(strip," \\end{center}");
      putc('\n',strip);  rownum++;
      }
    else			//InputPS.sty
      {
      fprintf(strip,"\\begin{forcewidth}");
      fprintf(strip,"{\\textwidth}\n"); rownum++;
      fprintf(strip," \\begin{center}");
      fprintf(strip,"\\InputPS{\\FigDir/%s.eps}",NewFilename() );
      putc('\n',strip);  rownum++;
      fprintf(strip," \\end{center}");
      putc('\n',strip);  rownum++;
      fprintf(strip,"\\end{forcewidth}\n");  rownum++;
      }

    InputPS |= 1;		//mark style as used
    Finalise_Conversion(this);
    RetVal++;
    }

  return(RetVal);
}


/* WMF files usually have special header before data. */

typedef struct _WmfSpecialHeader
{
  DWORD Key;           /* Magic number (always 9AC6CDD7h) */
  WORD  Handle;        /* Metafile HANDLE number (always 0) */
  SWORD Left;          /* Left coordinate in metafile units */
  SWORD Top;           /* Top coordinate in metafile units */
  SWORD Right;         /* Right coordinate in metafile units */
  SWORD Bottom;        /* Bottom coordinate in metafile units */
  WORD  Inch;          /* Number of metafile units per inch */
  DWORD Reserved;      /* Reserved (always 0) */
  WORD  Checksum;      /* Checksum value for previous 10 WORDs */
} WMFSPECIAL;


static inline long LoadWMFSpecial(FILE *f, _WmfSpecialHeader & WSH)
{
#if defined(__PackedStructures__)
return(fread(&WSH,1,sizeof(_WmfSpecialHeader),f));
#else
return(loadstruct(f,"dwwwwwwdw",
      &WSH.Key, &WSH.Handle,
      &WSH.Left, &WSH.Top, &WSH.Right, &WSH.Bottom,
      &WSH.Inch, &WSH.Reserved, &WSH.Checksum));
#endif
}


class TconvertedPass1_WMFspec: public TconvertedPass1_WMF
     {
public:
     virtual int Convert_first_pass(void);
     };


int TconvertedPass1_WMFspec::Convert_first_pass(void)
{
#ifdef DEBUG
  fprintf(log,"\n#Convert_pass1_WMF() ");fflush(log);
#endif

  {
  WMFSPECIAL hdr;
  if(LoadWMFSpecial(wpd,hdr)!=22) return -1;
  if(hdr.Key!=0x9AC6CDD7) return -2;

  DWORD Checksum = hdr.Key;
  Checksum ^= hdr.Handle;
  Checksum ^= hdr.Left;
  Checksum ^= hdr.Top;
  Checksum ^= hdr.Right;
  Checksum ^= hdr.Bottom;
  Checksum ^= hdr.Inch;
  Checksum ^= hdr.Reserved;

  Checksum = (Checksum & 0x0000FFFFUL) ^ ((Checksum & 0xFFFF0000UL) >> 16);
  if(Checksum!=hdr.Checksum)
    {
    if(err != NULL)
      {      
      fprintf(err, _("\nNote: Invalid WMF CRC %X, expected %X."),hdr.Checksum,Checksum);
      }
    }
  }

 return this->TconvertedPass1_WMF::Convert_first_pass();
}

/*Register translators here*/
TconvertedPass1 *Factory_WMFspec(void) {return new TconvertedPass1_WMFspec;}
FFormatTranslator FormatWMFWrapperSpecial("WMFspecial",Factory_WMFspec);


/*--------------------End of PASS1_WMF--------------------*/
